@@ -22,18 +22,63 @@ from trytond.modules.stock_package_shipping.exceptions import (
 from .configuration import get_client, SHIPMENT_SERVICE
 from .exceptions import DPDError
 
+TRACKING_URL = 'https://tracking.dpd.de/status/%(code)s/parcel/%(reference)s'
 
-class ShipmentOut(metaclass=PoolMeta):
-    __name__ = 'stock.shipment.out'
+
+class Package(metaclass=PoolMeta):
+    __name__ = 'stock.package'
+
+    def get_shipping_tracking_url(self, name):
+        pool = Pool()
+        ShipmentOut = pool.get('stock.shipment.out')
+        ShipmentInReturn = pool.get('stock.shipment.in.return')
+        url = super().get_shipping_tracking_url(name)
+        if (self.shipping_reference
+                and self.shipment
+                and self.shipment.id >= 0
+                and self.shipment.carrier
+                and self.shipment.carrier.shipping_service == 'dpd'):
+            party = address = None
+            if isinstance(self.shipment, ShipmentOut):
+                party = self.shipment.customer
+                address = self.shipment.delivery_address
+            elif isinstance(self.shipment, ShipmentInReturn):
+                party = self.shipment.supplier
+                address = self.shipment.delivery_address
+            if party and party.lang:
+                lang_code = party.lang.code
+            else:
+                lang_code = Transaction().language
+            if address and address.country:
+                code = '_'.join(
+                    (lang_code.split('_')[0], address.country.code))
+            else:
+                code = lang_code
+            url = TRACKING_URL % {
+                'code': code,
+                'reference': self.shipping_reference,
+                }
+        return url
+
+
+class ShippingDPDMixin:
 
     def validate_packing_dpd(self):
-        warehouse_address = self.warehouse.address
-        if not warehouse_address:
+        warehouse = self.shipping_warehouse
+        if not warehouse.address:
             raise PackingValidationError(
                 gettext('stock_package_shipping_dpd'
                     '.msg_warehouse_address_required',
                     shipment=self.rec_name,
-                    warehouse=self.warehouse.rec_name))
+                    warehouse=warehouse.rec_name))
+
+
+class ShipmentOut(ShippingDPDMixin, metaclass=PoolMeta):
+    __name__ = 'stock.shipment.out'
+
+
+class ShipmentInReturn(ShippingDPDMixin, metaclass=PoolMeta):
+    __name__ = 'stock.shipment.in.return'
 
 
 class CreateShipping(metaclass=PoolMeta):
@@ -43,12 +88,8 @@ class CreateShipping(metaclass=PoolMeta):
         'stock_package_shipping_dpd.act_create_shipping_dpd_wizard')
 
     def transition_start(self):
-        pool = Pool()
-        ShipmentOut = pool.get('stock.shipment.out')
-
-        shipment = ShipmentOut(Transaction().context['active_id'])
         next_state = super(CreateShipping, self).transition_start()
-        if shipment.carrier.shipping_service == 'dpd':
+        if self.record.carrier.shipping_service == 'dpd':
             next_state = 'dpd'
         return next_state
 
@@ -69,10 +110,9 @@ class CreateDPDShipping(Wizard):
 
     def transition_start(self):
         pool = Pool()
-        ShipmentOut = pool.get('stock.shipment.out')
         Package = pool.get('stock.package')
 
-        shipment = ShipmentOut(Transaction().context['active_id'])
+        shipment = self.record
         if shipment.reference:
             raise AccessError(
                 gettext('stock_package_shipping_dpd'
@@ -105,12 +145,13 @@ class CreateDPDShipping(Wizard):
                         })
                 break
             except Fault as e:
-                tag = etree.QName(e.detail[0].tag)
-                if tag.localname == 'authenticationFault':
-                    count += 1
-                    credential.update_token()
-                else:
-                    raise
+                if e.detail:
+                    tag = etree.QName(e.detail[0].tag)
+                    if tag.localname == 'authenticationFault':
+                        count += 1
+                        credential.update_token()
+                        continue
+                raise
         else:
             raise DPDError(
                 gettext('stock_package_shipping_dpd.msg_dpd_login_error',
@@ -138,7 +179,7 @@ class CreateDPDShipping(Wizard):
         for package, label, parcel in zip_longest(packages, labels, parcels):
             package.shipping_label = fields.Binary.cast(label.getvalue())
             package.shipping_reference = parcel.parcelLabelNumber
-        Package.save(shipment.root_packages)
+        Package.save(packages)
         shipment.save()
 
         return 'end'
@@ -159,8 +200,8 @@ class CreateDPDShipping(Wizard):
 
     def get_print_options(self, shipment):
         return {
-            'printerLanguage': 'PDF',
-            'paperFormat': 'A6',
+            'printerLanguage': shipment.carrier.dpd_printer_language,
+            'paperFormat': shipment.carrier.dpd_paper_format,
             }
 
     def shipping_party(self, party, address):
@@ -169,7 +210,7 @@ class CreateDPDShipping(Wizard):
             'name2': '',
             'street': ' '.join((address.street or '').splitlines())[:35],
             'country': address.country.code if address.country else '',
-            'zipCode': address.zip[:9],
+            'zipCode': address.postal_code[:9],
             'city': address.city[:35],
             }
         if party.full_name != address.party_full_name:
@@ -181,10 +222,10 @@ class CreateDPDShipping(Wizard):
                 phone = mechanism.value
             if mechanism.type == 'email' and not email:
                 email = mechanism.value
-        if phone:
-            shipping_party['phone'] = phone[:30]
-        if email:
-            shipping_party['email'] = email[:50]
+        if len(phone) <= 30:
+            shipping_party['phone'] = phone
+        if len(email) <= 50:
+            shipping_party['email'] = email
 
         return shipping_party
 
@@ -203,15 +244,15 @@ class CreateDPDShipping(Wizard):
             if weight < 1000000000:
                 parcel['weight'] = weight
 
-        if (package.type.length is not None
-                and package.type.width is not None
-                and package.type.height is not None):
+        if (package.length is not None
+                and package.width is not None
+                and package.height is not None):
             length = UoM.compute_qty(
-                package.type.length_uom, package.type.length, cm)
+                package.length_uom, package.length, cm)
             width = UoM.compute_qty(
-                package.type.width_uom, package.type.width, cm)
+                package.width_uom, package.width, cm)
             height = UoM.compute_qty(
-                package.type.height_uom, package.type.height, cm)
+                package.height_uom, package.height, cm)
             if length < 1000 and width < 1000 and height < 1000:
                 parcel['volume'] = int(
                     '%03i%03i%03i' % (length, width, height))
@@ -223,11 +264,12 @@ class CreateDPDShipping(Wizard):
             'generalShipmentData': {
                 'identificationNumber': shipment.number,
                 'sendingDepot': credential.depot,
-                'product': 'CL',
+                'product': shipment.carrier.dpd_product,
                 'sender': self.shipping_party(
-                    shipment.company.party, shipment.warehouse.address),
+                    shipment.company.party,
+                    shipment.shipping_warehouse.address),
                 'recipient': self.shipping_party(
-                    shipment.customer, shipment.delivery_address),
+                    shipment.shipping_to, shipment.shipping_to_address),
                 },
             'parcels': [self.get_parcel(p) for p in packages],
             'productAndServiceData': {
